@@ -29,6 +29,7 @@
 #include <gio/gio.h>
 #include <libgsystemservice/service.h>
 #include <locale.h>
+#include <polkit/polkit.h>
 #include <stdio.h>
 #include <systemd/sd-daemon.h>
 
@@ -44,9 +45,26 @@
  *
  * It follows the implementation recommendations in `man 7 daemon`.
  *
+ * Since 0.2.0, it registers a D-Bus object at `/org/freedesktop/Debugging`
+ * which exposes controls for debugging the service. Requests to these controls
+ * from peers must be authorized. The default policy is to check against polkit
+ * using the action ID set as #GssService:debug-controller-action-id, if set.
+ *
+ * If #GssService:debug-controller-action-id is not set, requests to the debug
+ * object will be accepted unconditionally if the service is running on the
+ * D-Bus session bus, and rejected unconditionally if the service is running on
+ * the system bus.
+ *
  * Since: 0.1.0
  */
 
+/* These require polkit 0.114 */
+#ifndef HAVE_POLKIT_114
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (PolkitAuthority, g_object_unref)
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (PolkitSubject, g_object_unref)
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (PolkitAuthorizationResult, g_object_unref)
+#endif  /* !HAVE_POLKIT_114 */
+
 /* These errors do not need to be registered with
  * g_dbus_error_register_error_domain() as they never go over the bus. */
 G_DEFINE_QUARK (GssServiceError, gss_service_error)
@@ -101,6 +119,10 @@ typedef struct
   guint inactivity_timeout_ms;  /* 0 indicates no timeout */
   GSource *inactivity_timeout_source;  /* (owned) (nullable) */
   guint hold_count;
+
+  GDebugController *debug_controller;  /* (owned) (nullable) */
+  gulong debug_controller_authorize_id;
+  gchar *debug_controller_action_id;  /* (owned) (nullable) */
 } GssServicePrivate;
 
 typedef enum
@@ -112,6 +134,8 @@ typedef enum
   PROP_SERVICE_ID,
   PROP_INACTIVITY_TIMEOUT,
   PROP_ALLOW_ROOT,
+  PROP_DEBUG_CONTROLLER_ACTION_ID,
+  PROP_DEBUG_CONTROLLER,
 } GssServiceProperty;
 
 G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (GssService, gss_service, G_TYPE_OBJECT)
@@ -120,7 +144,7 @@ static void
 gss_service_class_init (GssServiceClass *klass)
 {
   GObjectClass *object_class = (GObjectClass *) klass;
-  GParamSpec *props[PROP_ALLOW_ROOT + 1] = { NULL, };
+  GParamSpec *props[PROP_DEBUG_CONTROLLER + 1] = { NULL, };
 
   object_class->dispose = gss_service_dispose;
   object_class->get_property = gss_service_get_property;
@@ -207,7 +231,7 @@ gss_service_class_init (GssServiceClass *klass)
    */
   props[PROP_SERVICE_ID] =
       g_param_spec_string ("service-id", "Service ID",
-                           "The ID of the service, which must be well-known "
+                           "The ID of the service, which must be a well-known "
                            "D-Bus name to uniquely identify the service.",
                            NULL,
                            G_PARAM_READWRITE |
@@ -250,6 +274,46 @@ gss_service_class_init (GssServiceClass *klass)
                             G_PARAM_CONSTRUCT_ONLY |
                             G_PARAM_STATIC_STRINGS);
 
+  /**
+   * GssService:debug-controller-action-id: (nullable)
+   *
+   * The ID of a polkit action to check for authorization when a peer requests
+   * to change the debug settings.
+   *
+   * polkit will be queried with this action ID and the peer as the subject.
+   *
+   * If this is %NULL, the default security policy will be applied, as
+   * documented in #GssService.
+   *
+   * Since: 0.2.0
+   */
+  props[PROP_DEBUG_CONTROLLER_ACTION_ID] =
+      g_param_spec_string ("debug-controller-action-id",
+                           "Debug Controller Polkit Action ID",
+                           "The ID of a polkit action to check for "
+                           "authorization when a peer requests to change the "
+                           "debug settings.",
+                           NULL,
+                           G_PARAM_READWRITE |
+                           G_PARAM_CONSTRUCT_ONLY |
+                           G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GssService:debug-controller: (nullable)
+   *
+   * The debug controller used for the service. This exposes information about
+   * debug settings, such as whether debug output is enabled.
+   *
+   * Since: 0.2.0
+   */
+  props[PROP_DEBUG_CONTROLLER] =
+      g_param_spec_object ("debug-controller",
+                           "Debug Controller",
+                           "The debug controller used for the service.",
+                           G_TYPE_DEBUG_CONTROLLER,
+                           G_PARAM_READABLE |
+                           G_PARAM_STATIC_STRINGS);
+
   g_object_class_install_properties (object_class, G_N_ELEMENTS (props), props);
 }
 
@@ -292,6 +356,9 @@ gss_service_dispose (GObject *object)
   g_clear_pointer (&priv->summary, g_free);
   g_clear_pointer (&priv->service_id, g_free);
   g_clear_error (&priv->run_error);
+  g_clear_signal_handler (&priv->debug_controller_authorize_id, priv->debug_controller);
+  g_clear_object (&priv->debug_controller);
+  g_clear_pointer (&priv->debug_controller_action_id, g_free);
   g_clear_object (&priv->connection);
   g_clear_pointer (&priv->context, g_main_context_unref);
 
@@ -331,6 +398,12 @@ gss_service_get_property (GObject    *object,
     case PROP_ALLOW_ROOT:
       g_value_set_boolean (value, priv->allow_root);
       break;
+    case PROP_DEBUG_CONTROLLER_ACTION_ID:
+      g_value_set_string (value, priv->debug_controller_action_id);
+      break;
+    case PROP_DEBUG_CONTROLLER:
+      g_value_set_object (value, priv->debug_controller);
+      break;
     default:
       g_assert_not_reached ();
     }
@@ -378,6 +451,14 @@ gss_service_set_property (GObject      *object,
       /* Construct only. */
       priv->allow_root = g_value_get_boolean (value);
       break;
+    case PROP_DEBUG_CONTROLLER_ACTION_ID:
+      /* Construct only. */
+      g_assert (priv->debug_controller_action_id == NULL);
+      priv->debug_controller_action_id = g_value_dup_string (value);
+      break;
+    case PROP_DEBUG_CONTROLLER:
+      /* Read only */
+      G_GNUC_FALLTHROUGH;
     default:
       g_assert_not_reached ();
     }
@@ -554,6 +635,85 @@ check_for_early_exit (GssService  *self,
   return TRUE;
 }
 
+static gboolean
+debug_controller_authorize_cb (GDebugControllerDBus  *debug_controller,
+                               GDBusMethodInvocation *invocation,
+                               gpointer               user_data)
+{
+  GssService *self = GSS_SERVICE (user_data);
+  GssServicePrivate *priv = gss_service_get_instance_private (self);
+
+  /* Implement a security policy for allowing debug output to be enabled at
+   * runtime.
+   *
+   * If #GssService:debug-controller-action-id is set, polkit will be queried
+   * using that action ID, and its response will determine whether debug output
+   * can be enabled.
+   *
+   * Otherwise, if #GssService:bus-type is %G_BUS_TYPE_SESSION, enabling debug
+   * output will be allowed without authorization.
+   *
+   * Otherwise, enabling debug output will be denied.
+   *
+   * If services require more control over the security policy, #GssService will
+   * have to be extended to allow the service to provide their own
+   * #GDebugController, which they have hooked up to an appropriate security
+   * policy.
+   */
+  if (priv->debug_controller_action_id != NULL)
+    {
+      g_autoptr(PolkitAuthority) authority = NULL;
+      g_autoptr(PolkitSubject) subject = NULL;
+      g_autoptr(PolkitAuthorizationResult) auth_result = NULL;
+      g_autoptr(GError) local_error = NULL;
+      GDBusMessage *message;
+      GDBusMessageFlags message_flags;
+      PolkitCheckAuthorizationFlags flags = POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE;
+
+      message = g_dbus_method_invocation_get_message (invocation);
+      message_flags = g_dbus_message_get_flags (message);
+
+      authority = polkit_authority_get_sync (NULL, &local_error);
+      if (authority == NULL)
+        {
+          g_warning ("Failed to get polkit authority: %s", local_error->message);
+          return FALSE;
+        }
+
+      if (message_flags & G_DBUS_MESSAGE_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION)
+        flags |= POLKIT_CHECK_AUTHORIZATION_FLAGS_ALLOW_USER_INTERACTION;
+
+      subject = polkit_system_bus_name_new (g_dbus_method_invocation_get_sender (invocation));
+
+      auth_result = polkit_authority_check_authorization_sync (authority,
+                                                               subject,
+                                                               priv->debug_controller_action_id,
+                                                               NULL,
+                                                               flags,
+                                                               NULL,
+                                                               &local_error);
+      if (auth_result == NULL)
+        {
+          g_warning ("Failed to get check polkit authorization: %s", local_error->message);
+          return FALSE;
+        }
+
+      return polkit_authorization_result_get_is_authorized (auth_result);
+    }
+  else if (priv->bus_type == G_BUS_TYPE_SESSION)
+    {
+      /* All peers on the session bus are at the same privilege level anyway. */
+      return TRUE;
+    }
+  else
+    {
+      /* Peers on the system bus (or other buses) may have different privilege
+       * levels, so can’t be trusted to unconditionally enable debug mode on
+       * services. */
+      return FALSE;
+    }
+}
+
 /**
  * gss_service_run:
  * @self: a #GssService
@@ -722,6 +882,22 @@ gss_service_run (GssService  *self,
       return;
     }
 
+  /* Set up the debug controller. */
+  priv->debug_controller = G_DEBUG_CONTROLLER (g_debug_controller_dbus_new (priv->connection, NULL, &child_error));
+  if (priv->debug_controller == NULL)
+    {
+      g_set_error (error, GSS_SERVICE_ERROR, GSS_SERVICE_ERROR_NAME_UNAVAILABLE,
+                   _("Could not register debug controller on bus: %s"),
+                   child_error->message);
+      gss_service_release (self);
+      return;
+    }
+
+  priv->debug_controller_authorize_id = g_signal_connect (priv->debug_controller,
+                                                          "authorize",
+                                                          G_CALLBACK (debug_controller_authorize_cb),
+                                                          self);
+
   /* Start up. */
   g_autoptr(GAsyncResult) startup_result = NULL;
   g_assert (service_class->startup_async != NULL &&
@@ -927,6 +1103,25 @@ gss_service_set_inactivity_timeout (GssService *self,
   maybe_schedule_inactivity_timeout (self);
 }
 
+/**
+ * gss_service_get_debug_controller:
+ * @self: a #GssService
+ *
+ * Get the value of #GssService:debug-controller.
+ *
+ * Returns: (transfer none) (nullable): the service’s debug controller, or
+ *   %NULL if none is available
+ * Since: 0.2.0
+ */
+GDebugController *
+gss_service_get_debug_controller (GssService *self)
+{
+  g_return_val_if_fail (GSS_IS_SERVICE (self), NULL);
+
+  GssServicePrivate *priv = gss_service_get_instance_private (self);
+  return priv->debug_controller;
+}
+
 /**
  * gss_service_hold:
  * @self: a #GssService
